@@ -11,7 +11,8 @@ def fancyopts(args, options, state):
         map['-'+s] = map['--'+l] = pl
         state[pl] = d
         dt[pl] = type(d)
-        if not d is None and not callable(d):
+        if (d is not None and d is not True and d is not False and
+            not callable(d)):
             if s: s += ':'
             if l: l += '='
         if s: short = short + s
@@ -20,11 +21,12 @@ def fancyopts(args, options, state):
     opts, args = getopt.getopt(args, short, long)
 
     for opt, arg in opts:
-        if dt[map[opt]] is type(fancyopts): state[map[opt]](state,map[opt],arg)
+        if dt[map[opt]] is type(fancyopts): state[map[opt]](state, map[opt], arg)
         elif dt[map[opt]] is type(1): state[map[opt]] = int(arg)
         elif dt[map[opt]] is type(''): state[map[opt]] = arg
         elif dt[map[opt]] is type([]): state[map[opt]].append(arg)
-        elif dt[map[opt]] is type(None): state[map[opt]] = 1
+        elif dt[map[opt]] is type(None): state[map[opt]] = True
+        elif dt[map[opt]] is type(False): state[map[opt]] = True
 
     return args
 
